import numpy as np
import pandas as pd
from sklearn.decomposition import PCA
from vtk import vtkPolyData

from abm_shape_collection.construct_mesh_from_coeffs import construct_mesh_from_coeffs


def construct_mesh_from_points(
    pca: PCA,
    points: np.ndarray,
    feature_names: list[str],
    order: int,
    prefix: str = "",
    suffix: str = "",
) -> vtkPolyData:
    """Constructs mesh given PCA transformation points."""
    coeffs = pd.Series(pca.inverse_transform(points), index=feature_names)
    return construct_mesh_from_coeffs(coeffs, order, prefix, suffix)



import numpy as np
from aicsshparam import shtools
from vtk import vtkPolyData
from vtk.util import numpy_support


def construct_mesh_from_array(array: np.ndarray, reference: np.ndarray) -> vtkPolyData:
    _, angle = shtools.align_image_2d(image=reference)
    aligned_reference = shtools.apply_image_alignment_2d(reference, angle).squeeze()
    aligned_array = shtools.apply_image_alignment_2d(array, angle).squeeze()

    mesh, _, _ = shtools.get_mesh_from_image(
        image=aligned_array, lcc=False, translate_to_origin=False
    )

    centroid = np.argwhere(aligned_reference == 1).mean(axis=0)[[2, 1, 0]]
    coords = numpy_support.vtk_to_numpy(mesh.GetPoints().GetData()) - centroid
    mesh.GetPoints().SetData(numpy_support.numpy_to_vtk(coords))

    return mesh



import numpy as np
import pandas as pd
from sklearn.decomposition import PCA

from abm_shape_collection.calculate_size_stats import get_ks_statistic


def calculate_shape_stats(
    pca: PCA,
    data: pd.DataFrame,
    ref_data: pd.DataFrame,
    components: int,
    include_ticks: bool = False,
    include_samples: bool = False,
    sample_reps: int = 1,
    sample_size: int = 1,
) -> pd.DataFrame:
    all_stats = []

    columns = list(ref_data.filter(like="shcoeffs").columns)
    ref_data_transform = pca.transform(ref_data[columns].values)

    for component in range(components):
        stats = calculate_shape_stats_for_all(pca, data, ref_data_transform, component, columns)

        if include_ticks:
            stats = stats + calculate_shape_stats_for_ticks(
                pca, data, ref_data_transform, component, columns
            )

        if include_samples:
            stats = stats + calculate_shape_stats_for_samples(
                pca, data, ref_data_transform, component, columns, sample_reps, sample_size
            )

        stats_df = pd.DataFrame(stats)
        stats_df["FEATURE"] = f"PC_{component + 1}"
        all_stats.append(stats_df)

    all_stats_df = pd.concat(all_stats).astype({"N": int})

    return all_stats_df


def calculate_shape_stats_for_all(
    pca: PCA, data: pd.DataFrame, ref_data: np.ndarray, component: int, columns: list[str]
) -> list[dict]:
    ref_values = ref_data[:, component]
    values = pca.transform(data[columns].values)[:, component]

    ks_stats = get_ks_statistic(values, ref_values)
    ks_stats.update({"N": len(values)})

    return [ks_stats]


def calculate_shape_stats_for_ticks(
    pca: PCA, data: pd.DataFrame, ref_data: np.ndarray, component: int, columns: list[str]
) -> list[dict]:
    ref_values = ref_data[:, component]
    tick_stats = []

    for tick, tick_data in data.groupby("TICK"):
        tick_values = pca.transform(tick_data[columns].values)[:, component]

        tick_ks_stats = get_ks_statistic(tick_values, ref_values)
        tick_ks_stats.update({"TICK": tick, "N": len(tick_values)})

        tick_stats.append(tick_ks_stats)

    return tick_stats


def calculate_shape_stats_for_samples(
    pca: PCA,
    data: pd.DataFrame,
    ref_data: np.ndarray,
    component: int,
    columns: list[str],
    sample_reps: int,
    sample_size: int,
) -> list[dict]:
    ref_values = ref_data[:, component]
    sample_stats = []

    for sample in range(sample_reps):
        sample_values = pca.transform(
            data.sample(frac=1, random_state=sample)
            .groupby("TICK")
            .head(sample_size)[columns]
            .values
        )[:, component]

        sample_ks_stats = get_ks_statistic(sample_values, ref_values)
        sample_ks_stats.update({"SAMPLE": sample, "N": len(sample_values)})

        sample_stats.append(sample_ks_stats)

    return sample_stats



import numpy as np
import pandas as pd
from sklearn.decomposition import PCA

from abm_shape_collection.construct_mesh_from_points import construct_mesh_from_points
from abm_shape_collection.extract_mesh_projections import extract_mesh_projections


def extract_shape_modes(
    pca: PCA, data: pd.DataFrame, components: int, regions: list[str], order: int, delta: float
) -> dict:
    # Transform data into shape mode space.
    features = data.filter(like="shcoeffs").columns
    transform = pca.transform(data[features].values)

    # Calculate transformed means and standard deviations.
    means = transform.mean(axis=0)
    stds = transform.std(axis=0, ddof=1)

    # Create bins.
    map_points = np.arange(-2, 2.5, delta)
    bin_edges = [-np.inf] + [point + delta / 2 for point in map_points[:-1]] + [np.inf]
    transform_binned = np.digitize(transform / stds, bin_edges)

    # Calculate offsets.
    offsets = {
        region: calculate_region_offsets(data, region) for region in regions if region != "DEFAULT"
    }

    # Initialize output dictionary.
    shape_modes: dict[str, list] = {}

    for region in regions:
        region_shape_modes = []

        suffix = f".{region}" if region != "DEFAULT" else ""
        offsets = calculate_region_offsets(data, region)

        for component in range(components):
            point_vector = np.zeros((components))

            for point in map_points:
                point_bin = np.digitize(point, bin_edges)
                point_vector[component] = point

                vector = means + np.multiply(stds, point_vector)
                indices = transform_binned[:, component] == point_bin

                mesh = construct_mesh_from_points(pca, vector, features, order, suffix=suffix)

                if region == "DEFAULT":
                    offset = None
                else:
                    offset = (
                        offsets["x"][indices].mean(),
                        offsets["y"][indices].mean(),
                        offsets["z"][indices].mean(),
                    )

                region_shape_modes.append(
                    {
                        "mode": component + 1,
                        "point": point,
                        "projections": extract_mesh_projections(mesh, extents=False, offset=offset),
                    }
                )

        shape_modes[region] = region_shape_modes

    return shape_modes


def calculate_region_offsets(data: pd.DataFrame, region: str) -> dict:
    if region == "DEFAULT":
        return {}

    x_deltas = data[f"CENTER_X.{region}"].values - data["CENTER_X"].values
    y_deltas = data[f"CENTER_Y.{region}"].values - data["CENTER_Y"].values
    z_deltas = data[f"CENTER_Z.{region}"].values - data["CENTER_Z"].values
    angles = data["angle"].values * np.pi / 180.0

    sin_angles = np.sin(angles)
    cos_angles = np.cos(angles)

    return {
        "x": x_deltas * cos_angles - y_deltas * sin_angles,
        "y": x_deltas * sin_angles + y_deltas * cos_angles,
        "z": z_deltas,
    }



from typing import Optional, Union

import trimesh
from vtk import vtkPolyData

from abm_shape_collection.extract_mesh_projections import convert_vtk_to_trimesh


def extract_mesh_wireframe(
    mesh: Union[vtkPolyData, trimesh.Trimesh], offset: Optional[tuple[float, float, float]] = None
) -> list[list[list[float]]]:
    if isinstance(mesh, vtkPolyData):
        mesh = convert_vtk_to_trimesh(mesh)

    if offset is not None:
        mesh.apply_translation(offset)

    wireframe = [[list(mesh.vertices[a]), list(mesh.vertices[b])] for a, b in mesh.edges]

    return wireframe



import tempfile
from typing import Optional, Union

import numpy as np
import trimesh
from vtk import vtkPLYWriter, vtkPolyData

PROJECTIONS: list[tuple[str, tuple[int, int, int], int]] = [
    ("side1", (0, 1, 0), 1),
    ("side2", (1, 0, 0), 0),
    ("top", (0, 0, 1), 2),
]


def extract_mesh_projections(
    mesh: Union[vtkPolyData, trimesh.Trimesh],
    slices: bool = True,
    extents: bool = True,
    offset: Optional[tuple[float, float, float]] = None,
) -> dict:
    if isinstance(mesh, vtkPolyData):
        mesh = convert_vtk_to_trimesh(mesh)

    if offset is not None:
        mesh.apply_translation(offset)

    projections: dict[
        str, Union[list[list[list[float]]], dict[float, list[list[list[float]]]]]
    ] = {}

    if slices:
        for projection, normal, _ in PROJECTIONS:
            projections[f"{projection}_slice"] = get_mesh_slice(mesh, normal)

    if extents:
        for projection, normal, index in PROJECTIONS:
            projections[f"{projection}_extent"] = get_mesh_extent(mesh, normal, index)

    return projections


def convert_vtk_to_trimesh(mesh: vtkPolyData) -> trimesh.Trimesh:
    with tempfile.NamedTemporaryFile() as temp:
        writer = vtkPLYWriter()
        writer.SetInputData(mesh)
        writer.SetFileTypeToASCII()
        writer.SetFileName(f"{temp.name}.ply")
        _ = writer.Write()
        mesh = trimesh.load(f"{temp.name}.ply")

    return mesh


def get_mesh_slice(mesh: trimesh.Trimesh, normal: tuple[int, int, int]) -> list[list[list[float]]]:
    """Get slice of mesh along plane for given normal as path points."""
    mesh_slice = mesh.section_multiplane((0, 0, 0), normal, [0])
    points = [[list(point) for point in entity] for entity in mesh_slice[0].discrete]
    return points


def get_mesh_extent(
    mesh: trimesh.Trimesh, normal: tuple[int, int, int], index: int
) -> dict[float, list[list[list[float]]]]:
    """Get extent of mesh along plane for given normal as path points."""
    layers = int(mesh.extents[index] + 2)
    plane_indices = list(np.arange(-layers, layers + 1, 0.5))
    mesh_extents = mesh.section_multiplane((0, 0, 0), normal, plane_indices)
    points = {
        index: [[list(point) for point in entity] for entity in mesh_extent.discrete]
        for mesh_extent, index in zip(mesh_extents, plane_indices)
        if mesh_extent is not None
    }
    return points



import numpy as np
import pandas as pd
from aicsshparam import shtools
from vtk import vtkPolyData, vtkTransform, vtkTransformPolyDataFilter


def construct_mesh_from_coeffs(
    coeffs: pd.DataFrame,
    order: int,
    prefix: str = "",
    suffix: str = "",
    scale: float = 1.0,
) -> vtkPolyData:
    coeffs_map = np.zeros((2, order + 1, order + 1), dtype=np.float32)

    for l in range(order + 1):
        for m in range(order + 1):
            coeffs_map[0, l, m] = coeffs[f"{prefix}shcoeffs_L{l}M{m}C{suffix}"]
            coeffs_map[1, l, m] = coeffs[f"{prefix}shcoeffs_L{l}M{m}S{suffix}"]

    mesh, _ = shtools.get_reconstruction_from_coeffs(coeffs_map)

    if scale != 1:
        transform = vtkTransform()
        transform.Scale((scale, scale, scale))

        transform_filter = vtkTransformPolyDataFilter()
        transform_filter.SetInputData(mesh)
        transform_filter.SetTransform(transform)
        transform_filter.Update()

        mesh = transform_filter.GetOutput(0)

    return mesh



import numpy as np
import pandas as pd
from scipy.stats import ks_2samp


def calculate_size_stats(
    data: pd.DataFrame,
    ref_data: pd.DataFrame,
    regions: list[str],
    include_ticks: bool = False,
    include_samples: bool = False,
    sample_reps: int = 1,
    sample_size: int = 1,
) -> pd.DataFrame:
    all_stats = []

    features = [
        f"{feature}.{region}" if region != "DEFAULT" else feature
        for region in regions
        for feature in ["volume", "height"]
    ]

    for feature in features:
        stats = calculate_size_stats_for_all(data, ref_data, feature)

        if include_ticks:
            stats = stats + calculate_size_stats_for_ticks(data, ref_data, feature)

        if include_samples:
            stats = stats + calculate_size_stats_for_samples(
                data, ref_data, feature, sample_reps, sample_size
            )

        stats_df = pd.DataFrame(stats)
        stats_df["FEATURE"] = feature
        all_stats.append(stats_df)

    all_stats_df = pd.concat(all_stats).astype({"N": int})

    return all_stats_df


def calculate_size_stats_for_all(
    data: pd.DataFrame, ref_data: pd.DataFrame, feature: str
) -> list[dict]:
    ref_values = ref_data[feature].values
    values = data[feature].values

    ks_stats = get_ks_statistic(values, ref_values)
    ks_stats.update({"N": len(values)})

    return [ks_stats]


def calculate_size_stats_for_ticks(
    data: pd.DataFrame, ref_data: pd.DataFrame, feature: str
) -> list[dict]:
    ref_values = ref_data[feature].values
    tick_stats = []

    for tick, tick_data in data.groupby("TICK"):
        tick_values = tick_data[feature].values

        tick_ks_stats = get_ks_statistic(tick_values, ref_values)
        tick_ks_stats.update({"TICK": tick, "N": len(tick_values)})

        tick_stats.append(tick_ks_stats)

    return tick_stats


def calculate_size_stats_for_samples(
    data: pd.DataFrame, ref_data: pd.DataFrame, feature: str, sample_reps: int, sample_size: int
) -> list[dict]:
    ref_values = ref_data[feature].values
    sample_stats = []

    for sample in range(sample_reps):
        sample_values = (
            data.sample(frac=1, random_state=sample)
            .groupby("TICK")
            .head(sample_size)[feature]
            .values
        )

        sample_ks_stats = get_ks_statistic(sample_values, ref_values)
        sample_ks_stats.update({"SAMPLE": sample, "N": len(sample_values)})

        sample_stats.append(sample_ks_stats)

    return sample_stats


def get_ks_statistic(population_a: np.ndarray, population_b: np.ndarray) -> dict:
    ksresult = ks_2samp(population_a, population_b, mode="asymp")

    return {
        "KS_STATISTIC": ksresult.statistic,
        "KS_PVALUE": ksresult.pvalue,
    }



import numpy as np
from skimage.measure import regionprops_table


def get_shape_properties(array: np.ndarray, properties: list[str]) -> dict:
    flattened_array = array.max(axis=0)

    properties_dict = regionprops_table(flattened_array, properties=properties)
    props = {key: value[0] for key, value in properties_dict.items()}

    return props



import numpy as np
from aicsshparam import shparam, shtools


def get_shape_coefficients(array: np.ndarray, reference: np.ndarray, lmax: int) -> dict:
    _, angle = shtools.align_image_2d(image=reference)
    aligned_array = shtools.apply_image_alignment_2d(array, angle).squeeze()

    (coeffs, reconstructed_array), (_, _, downsampled_array, _) = shparam.get_shcoeffs(
        image=aligned_array, lmax=lmax, compute_lcc=False, alignment_2d=False
    )

    mse = shtools.get_reconstruction_error(downsampled_array, reconstructed_array)
    coeffs["angle"] = angle
    coeffs["mse"] = mse

    return coeffs



from prefect import task

from .calculate_shape_stats import calculate_shape_stats
from .calculate_size_stats import calculate_size_stats
from .construct_mesh_from_array import construct_mesh_from_array
from .construct_mesh_from_coeffs import construct_mesh_from_coeffs
from .construct_mesh_from_points import construct_mesh_from_points
from .extract_mesh_projections import extract_mesh_projections
from .extract_mesh_wireframe import extract_mesh_wireframe
from .extract_shape_modes import extract_shape_modes
from .fit_pca_model import fit_pca_model
from .get_shape_coefficients import get_shape_coefficients
from .get_shape_properties import get_shape_properties
from .make_voxels_array import make_voxels_array

calculate_shape_stats = task(calculate_shape_stats)
calculate_size_stats = task(calculate_size_stats)
construct_mesh_from_array = task(construct_mesh_from_array)
construct_mesh_from_coeffs = task(construct_mesh_from_coeffs)
construct_mesh_from_points = task(construct_mesh_from_points)
extract_mesh_projections = task(extract_mesh_projections)
extract_mesh_wireframe = task(extract_mesh_wireframe)
extract_shape_modes = task(extract_shape_modes)
fit_pca_model = task(fit_pca_model)
get_shape_coefficients = task(get_shape_coefficients)
get_shape_properties = task(get_shape_properties)
make_voxels_array = task(make_voxels_array)
